/* $Id: pgheader.c,v 1.3 2005/04/19 18:16:47 nanardon Exp $ */
#include <postgres.h>
#include <fmgr.h>
#include <utils/builtins.h>
#include <funcapi.h>

#include <rpm/rpmlib.h>
#include <rpm/header.h>
#include <rpm/rpmts.h>

#include "pgrpm.h"
#include "pgrpmlib.h"
#include "pgheader.h"

bytea *
headergetDatum(Header h)
{
    bytea * out;
    void * ptr = NULL;
    int size = headerSizeof(h, HEADER_MAGIC_NO);
    ptr = headerUnload(h);
    out = palloc(size + VARHDRSZ);
    VARATT_SIZEP(out) = size + VARHDRSZ;
    memcpy(VARDATA(out), ptr, size);
    _free(ptr);
    return out;
}

Header
Datumgetheader(bytea * in, int * size)
{
    Header h;
    if (size)
        *size = VARSIZE(in) - VARHDRSZ;
    h = headerLoad(VARDATA(in));
    return h;
}
    

PG_FUNCTION_INFO_V1(header_in);

Datum
header_in(PG_FUNCTION_ARGS)
{
    Datum * dest = NULL;
    
    *dest = DirectFunctionCall1(byteain, CStringGetDatum(PG_GETARG_CSTRING(0)));

    if (headerCheck(NULL, VARDATA(dest), VARSIZE(dest)-VARHDRSZ, NULL) != RPMRC_OK) {
        ereport(ERROR,
            (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
            errmsg("Data input are not a header")));
    }
    
    PG_RETURN_BYTEA_P(dest);
}

PG_FUNCTION_INFO_V1(header_out);

Datum
header_out(PG_FUNCTION_ARGS)
{
    PG_RETURN_CSTRING(DirectFunctionCall1(byteaout, PG_GETARG_DATUM(0)));
} 

PG_FUNCTION_INFO_V1(headerfromfile);

Datum
headerfromfile(PG_FUNCTION_ARGS)
{
    rpmts ts = rpmtsCreate();
    char * filename;
    FD_t fd;
    Header h = NULL;
    rpmRC rc = 0;
    bytea * out = NULL;
    
    filename = GET_STR(PG_GETARG_TEXT_P(0));
    if ((fd = Fopen(filename, "r"))) {
        rpmtsSetVSFlags(ts, _RPMVSF_NOPAYLOAD | _RPMVSF_NOSIGNATURES);
        
        rc = rpmReadPackageFile(ts, fd, filename, &h);
        if (rc == RPMRC_OK) {
            out = headergetDatum(h);
        } else {
            ereport(ERROR,
                (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
                errmsg("Can't get valid header from file")));
        }
    } else {
        ereport(ERROR,
            (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
             errmsg("Can't open file")));
    }
    Fclose(fd);
    ts = rpmtsFree(ts);
    h = headerFree(h);
    PG_RETURN_BYTEA_P(out);
}

PG_FUNCTION_INFO_V1(header_queryformat);

Datum
header_queryformat(PG_FUNCTION_ARGS)
{
    Header h = Datumgetheader(PG_GETARG_BYTEA_P_COPY(0), NULL);
    char * format = GET_STR(PG_GETARG_TEXT_P(1));
    char * res = NULL;
    text * tres = NULL;
    int size;
    res = headerSprintf(h, format, rpmTagTable, rpmHeaderFormats, NULL);
    if (res) {
        size = strlen(res) + VARHDRSZ;
        tres = palloc(size);
        VARATT_SIZEP(tres) = size;
        memcpy(VARDATA(tres), res, size - VARHDRSZ);
    }    
    _free(res); /* this malloc come from rpmlib */
    h = headerFree(h);
    if (tres)
        PG_RETURN_TEXT_P(tres);
    else
        PG_RETURN_NULL();
}

PG_FUNCTION_INFO_V1(header_query);

Datum
header_query(PG_FUNCTION_ARGS)
{
    FuncCallContext     *funcctx;
    int                  call_cntr;
    int                  max_calls;
    if (SRF_IS_FIRSTCALL()) {
        MemoryContext   oldcontext;
        Header h = Datumgetheader(PG_GETARG_BYTEA_P_COPY(0), NULL);
        int tag = PG_GETARG_INT32(1);
        int tagtype = 0;
        void *ret;
        Datum **fmtret = NULL;
        int count = 0;
        funcctx = SRF_FIRSTCALL_INIT();
        oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
        if (headerGetEntry(h, tag, &tagtype, &ret, &count)) {
            switch(tagtype) {
                case RPM_STRING_TYPE:
                {
                    int size;
                    fmtret = (Datum **) palloc(sizeof(Datum *));
                    size = strlen(ret) + VARHDRSZ;
                    fmtret[0] = palloc(size);
                    memcpy(VARDATA(fmtret[0]), ret, size - VARHDRSZ);
                    VARATT_SIZEP(fmtret[0]) = size;
                    funcctx->max_calls = 1;
                }
                break;
                case RPM_STRING_ARRAY_TYPE:
                {
                    int i;
                    char ** sptr;
                    fmtret = (Datum **) palloc(count * sizeof(Datum *));
                    sptr = (char **)ret;
                    for (i = 0; i < count; i ++) {
                        int size = strlen(sptr[i]) + VARHDRSZ;
                        fmtret[i] = palloc(size);
                        memcpy(VARDATA(fmtret[i]), sptr[i], size - VARHDRSZ);
                        VARATT_SIZEP(fmtret[i]) = size;
                    }
                    funcctx->max_calls = count;
                }
                break;
                case RPM_INT8_TYPE:
                case RPM_INT16_TYPE:
                case RPM_INT32_TYPE:
                {
                    int i;
                    int *r;
                    fmtret = (Datum **) palloc(count * sizeof(Datum *));
                    r = (int *)ret;
                    for (i = 0; i < count; i ++) {
                        char * fmt = palloc(16);
                        int size;
                        snprintf(fmt, 16, "%d", r[i]);
                        size = strlen(fmt) + VARHDRSZ;
                        fmtret[i] = palloc(size);
                        memcpy(VARDATA(fmtret[i]), fmt, size - VARHDRSZ);
                        VARATT_SIZEP(fmtret[i]) = size;
                        pfree(fmt);
                    }
                    funcctx->max_calls = count;
                }
                break;
                case RPM_BIN_TYPE:
                {
                    fmtret = (Datum **) palloc(sizeof(Datum *));
                    fmtret[0] = palloc(count * sizeof(char) + VARHDRSZ);
                    memcpy(VARDATA(fmtret[0]), ret, count);
                    VARATT_SIZEP(fmtret[0]) = count + VARHDRSZ;
                    funcctx->max_calls = 1;
                }
                break;
                default:
                break;
            }
            funcctx->user_fctx = fmtret;
        }
        headerFreeTag(h, ret, tagtype);
        headerFree(h);
        MemoryContextSwitchTo(oldcontext);
    }
    funcctx = SRF_PERCALL_SETUP();
    
    call_cntr = funcctx->call_cntr;
    max_calls = funcctx->max_calls;
    if (call_cntr < max_calls)
    {
        Datum       **values = funcctx->user_fctx;

        SRF_RETURN_NEXT(funcctx, (Datum) values[call_cntr]);

    }
    else    /* do when there is no more left */
    {
        SRF_RETURN_DONE(funcctx);
    }
}

